/*
   D1M-WX1_IoT-APRS.ino
   D1 Mini Weather Station (Solar)
   Posts to ThingSpeak & APRS-IS

   Set serial monitor to 115,200 baud

   3/14/2017 - release
   Copyright(c) 2017 Berger Engineering

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in all
   copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
*/

#include <Wire.h>                // I2C bus
#include <BME280I2C.h>           // [manager] by Tyler Glenn https://github.com/finitespace/BME280
#include <BH1750.h>              // https://github.com/claws/BH1750
#include <ESP8266WiFi.h>         // ESP8266 WiFi
#include <WiFiUdp.h>             // UDP for Net Time Protocol service

#include "config.h"              // Your configuration file

// *******************************************************
// ******************* GLOBALS ***************************
// *******************************************************
int LEDpin = 2;                                // for LED on ESP8266 chip
const float HPA_TO_INHG = 0.0295299830714;     // hPa (mb) to inHg pressure

// structure to hold sensor measurements in metric units
struct sensorData
{
  float stationPressure;
  float seaLevelPressure;
  float temperature;
  float humidity;
  unsigned int lightIntensity;
  float cellVoltage;
  long wifiRSSI;
} rawData;  // declare struct variables

// used in printToSerial() to select display units
const bool customaryUnits = false;
const bool metricUnits = true;

// *******************************************************
// ******************* DEFAULTS **************************
// *******************************************************
// do not change these defaults
const char APRS_DeviceName[] = "http://w4krl.com/iot-kits";
const char APRSver[] = "1.0";
const char APRS_SVR_PROMPT[] = "javAPRSSrvr";
char APRS_SVR_VERIFIED[] = "verified";
const long TO_LINE = 10000;    // used in wait4APRS()
// NTP Server:
static const char ntpServerName[] = "us.pool.ntp.org";

// ********* INSTANTIATE OBJECTS *************************

// Create instances of the sensor objects
BME280I2C myBME280;     // barometric pressure / temperature / humidity sensor
BH1750 myBH1750;        // light intensity sensor

// instantiate the WiFi connection
WiFiClient client;

WiFiUDP udp;
unsigned int localPort = 8888;  // local port for NTP UDP packets 2390, 123???

// *******************************************************
// ********************* SETUP ***************************
// *******************************************************
void setup()
{
  Serial.begin(115200);             // initialize the serial port
  myBME280.begin();                 // initialize BME280 pressure/temperature/humidity
  myBH1750.begin();                 // initialize BH1750 light sensor

  logonToRouter();                  // logon to local Wi-Fi
  udp.begin(localPort);             // NTP

  rawData = readSensors();          // load all sensor data into rawData struct

  printToSerialPort(rawData, customaryUnits);
  postToThingSpeak(rawData);
  postToAPRS(rawData);

  // all done, now go to sleep
  blinkLED(3);                      // notify the user
  Serial.println("Entering deep sleep");
  delay(200);                       // delay to let things settle

  ESP.deepSleep(SLEEP_INTERVAL * 1000000L, WAKE_RF_DEFAULT);
} //setup()

// *******************************************************
// ******************** LOOP *****************************
// *******************************************************
void loop()
{
  // everything is done in Setup()
} // loop()

// *******************************************************
// ****************** SUPPORTING FUNCTIONS ***************
// *******************************************************

// *******************************************************
// ********** read sensors into a sensor struct **********
// *******************************************************
sensorData readSensors()
{
  sensorData tempData;   // temporary variable to hold readings

  // Pressure Parameters: (float pressure, float temperature, float humidity, bool metric, int pressureUnit)
  int pressureUnit = 1;  // unit: 0 = Pa, 1 = hPa, 2 = Hg, 3 = atm, 4 = bar, 5 = torr, 6 = N/m^2, 7 = psi
  myBME280.read(tempData.stationPressure, tempData.temperature, tempData.humidity, true, pressureUnit);

  // calculater the Sea Level Pressure from the station pressure
  float SLP = calculateSeaLevelPressure(tempData.temperature, tempData.stationPressure, STATION_ELEV);
  tempData.seaLevelPressure = SLP;

  // read light level in lux
  tempData.lightIntensity = myBH1750.readLightLevel();

  // read analog voltage from the Analog to Digital Converter
  // on D1 Mini this is 0 - 1023 for voltages 0 to 3.2V
  // the D1M-WX1 has an external resistor to extend the range to 5.0 Volts
  // a fudgeFactor corrects for component variation
  float fudgeFactor = dmmVoltage / adcVoltage;
  tempData.cellVoltage = 5.0 * analogRead(A0) * fudgeFactor / 1023;

  // read the Wi-Fi signal strength
  tempData.wifiRSSI = WiFi.RSSI();

  return tempData;
} // readSensors()

// *******************************************************
// ********* print raw data to the serial port ***********
// *******************************************************
void printToSerialPort(sensorData dataRaw, bool units)
{
  // '\t' is the C++ escape sequence for tab
  // header line
  Serial.println("\tTemp:\tHumd:\tPsta:\tPsea:\tLux:\tVolt:");

  if (units == true) { // display metric units
    // first line - raw data
    Serial.print("RAW\t");
    Serial.print(dataRaw.temperature, 2);
    Serial.print("\t");
    Serial.print(dataRaw.humidity, 2);
    Serial.print("\t");
    Serial.print(dataRaw.stationPressure * HPA_TO_INHG, 2);
    Serial.print("\t");
    Serial.print(dataRaw.seaLevelPressure * HPA_TO_INHG, 2);
    Serial.print("\t");
    Serial.print(dataRaw.lightIntensity);
    Serial.print("\t");
    Serial.println(dataRaw.cellVoltage, 2);
  } else {           // display customary units
    // first line - raw data
    Serial.print("RAW\t");
    Serial.print(dataRaw.temperature * 1.8 + 32, 2);
    Serial.print("\t");
    Serial.print(dataRaw.humidity, 2);
    Serial.print("\t");
    Serial.print(dataRaw.stationPressure, 2);
    Serial.print("\t");
    Serial.print(dataRaw.seaLevelPressure, 2);
    Serial.print("\t");
    Serial.print(dataRaw.lightIntensity);
    Serial.print("\t");
    Serial.println(dataRaw.cellVoltage, 2);
  }
  // second line
  Serial.println("----------------------------------------------------");
} // printToSerialPort()

// *******************************************************
// *********** logon to your Wi-Fi router ****************
// *******************************************************
void logonToRouter()
{
  int count = 0;
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED)
  {
    count++;
    blinkLED(1);
    delay(250);
    Serial.print(".");
    if (count > 15) {
      Serial.print("Exit code: ");
      Serial.println(WiFi.status());
      ESP.reset();
    }
  }
  Serial.print("Connected w/IP: ");
  Serial.println(WiFi.localIP().toString());
} // logonToRouter()

// *******************************************************
// ************** post to APRS server ********************
// *******************************************************
void postToAPRS(sensorData dataRaw) {
  String dataString = "";
  if ( client.connect(APRS_SERVER, APRS_PORT) )
  {
    Serial.println("APRS Server connected");
    dataString = "user ";
    dataString += APRScallsign;
    dataString += " pass ";
    dataString += APRSpasscode;
    dataString += " vers APRSduino ";
    dataString += APRSver;
    client.println(dataString);   // send to APRS-IS
    Serial.println(dataString);   // print to serial monitor
    if ( wait4APRS(&client, APRS_SVR_VERIFIED, 8) )
    {
      Serial.println("APRS Login ok");

      APRSsendWeather(rawData);
      APRSsendTelemetry(rawData);

      client.stop();
      Serial.println("APRS Server disconnected");
    }
    else
    {
      Serial.println("APRS Login failed.");
    } // if wait4aprs
  } // if connect
} // postToAPRS()

// *******************************************************
// ************* post to ThingSpeak **********************
// *******************************************************
void postToThingSpeak(sensorData data)
{
  // assemble and post the data
  if (client.connect(IOT_SERVER, 80) == true)
  {
    // declare dataString as a String and initialize with the apiWriteKey
    String dataString = apiWriteKey;

    // cocatenate each field onto the end of dataString
    dataString += "&field1=";
    dataString += String(data.temperature);
    dataString += "&field2=";
    dataString += String(data.humidity);
    dataString += "&field3=";
    dataString += String(data.stationPressure);
    dataString += "&field4=";
    dataString += String(data.seaLevelPressure);
    dataString += "&field5=";
    dataString += String(data.lightIntensity);
    dataString += "&field6=";
    dataString += String(data.cellVoltage);
    dataString += "&field7=";
    dataString += String(data.wifiRSSI);
    //    dataString += "&field8=";   // spare field
    //    dataString += String();
    //    dataString += "&status=";   // unused status field
    //    dataString += sensorStatus;

    // find the number of characters in dataString
    String dataStringLength = String(dataString.length());

    // post the data to ThingSpeak
    client.println("POST /update HTTP/1.1");
    client.println("Host: api.thingspeak.com");
    client.println("Connection: close");
    client.println("X-THINGSPEAKAPIKEY: " + apiWriteKey);
    client.println("Content-Type: application/x-www-form-urlencoded");
    client.println("Content-Length: " + dataStringLength);
    client.println("");
    client.print(dataString);
  }
  client.stop();
} // postToThingSpeak()

// *******************************************************
// calculate relative sealevel pressure from absolute station pressure in hPa
// temperature in °C, elevation in m
// http://www.sandhurstweather.org.uk/barometric.pdf
// http://keisan.casio.com/exec/system/1224575267
// *******************************************************
float calculateSeaLevelPressure(float celsius, float stationPressure, float elevation)
{
  float slP = stationPressure / pow(2.718281828, -(elevation / ((273.15 + celsius) * 29.263)));
  return slP;
} // calculateSeaLevelPressure()

// *******************************************************
// ******** format callsign for APRS telemetry ***********
// *******************************************************
// max length with SSID is 9, for ex.: WA3YST-13
// this pads a short call with spaces to the right
String APRSpadCall(String callSign) {
  int len = callSign.length();
  String dataString = callSign;
  for (int i = len; i < 9; i++) {
    dataString += " ";
  }
  return dataString;
}  // APRSpadCall()

// *******************************************************
// ******** format temperature in Fahrenheit for APRS ****
// *******************************************************
String APRSpadTempF(float t)
{
  // convert C to F for APRS
  t = 1.8 * t + 32;
  String dataString = "";
  int i = t + 0.5;       // round to integer temperature
  if ( i < 0 ) {         // temperature is below zero
    dataString += "-";   // prefix with minus sign
    i = -i;              // make value positive
    if ( i > 99 )        // limit temperature to -99F
      i = 99;
    else if ( i < 10 )   // if between -1F and -9F
      dataString += "0"; // pad with leading zero
  } else {               // temperature is positive
    if ( i < 10 )        // pad with leading 00 between 0F and 9F
      dataString += "00";
    else if ( i < 100 )  // pad with leading 0 between 10F and 99F
      dataString += "0";
  }
  dataString += i;       // add padded string
  return dataString;
} // APRSpadTempF

// *******************************************************
// *************** format humidity for APRS **************
// *******************************************************
String APRSpadHumid(float h)
{
  String dataString = "";
  int i = h + 0.5;      // rounds up and converts to integer
  if ( i > 99 ) {
    i = 99;
  }
  if ( i < 10 ) {
    dataString += "0";
  }
  dataString += i;
  return dataString;
} // APRSpadHumid()

// *******************************************************
// ******** format LiPo cell voltage for APRS ************
// *******************************************************
String APRSpadVcell(float vcell)
{
  String dataString = "";
  int v = vcell * 255 / 5.0;  // convert voltage to 0-255
  if ( v < 10 ) {
    dataString += "00";
  } else if ( v < 100 ) {
    dataString += "0";
  }
  dataString += v;
  return dataString;
} // APRSpadVcell()

// *******************************************************
// ********** format light intensity for APRS ************
// *******************************************************
String APRSpadLightIntensity(long li)
{
  String dataString = "";
  unsigned int t = li / 256;
  if ( t < 10 ) {
    dataString += "00";
  } else if ( t < 100 ) {
    dataString += "0";
  }
  dataString += t;
  return dataString;
} // APRSpadLightIntensity()

// *******************************************************
// ************* format WiFi RSSI for APRS ***************
// *******************************************************
String APRSpadRSSI(float t)
{
  String dataString = "";
  int i = -t;
  if ( i < 10 )
    dataString += "00";
  else if ( i < 100 )
    dataString += "0";
  dataString += i;
  return dataString;
} // APRSpadRSSI()

// *******************************************************
// ********** format sea level pressure for APRS *********
// *******************************************************
String APRSpadBaro(float b)
{
  int i = b * 10;    // convert to 10th of millibars
  String dataString = "";
  if (i > 99999) {
    i = 99999;
    dataString += i;
  }
  if (i > 10000 && i <= 99999) {
    dataString += i;
  }
  if (i < 10000 && i >= 1000) {
    dataString += "01";
  }
  if (i < 1000 && i > 0) {
    dataString += "00";
    dataString += i;
  }
  if (i < 0) {
    dataString += "00000";
  }
  return dataString;
} // APRSpadBaro()

// *******************************************************
// ************ send weather data to APRS-IS *************
// *******************************************************
// page 65 http://www.aprs.org/doc/APRS101.PDF
void APRSsendWeather(sensorData dataRaw) {
  String dataString = APRScallsign;
  dataString += ">APRS,TCPIP*:!";
  dataString += APRSlocation;
  dataString += "_.../...g...";   // no wind data
  dataString += "t";              // temperature
  dataString += APRSpadTempF(dataRaw.temperature);
  dataString += "p...P...";       // no rainfall data
  dataString += "h";              // humidity
  dataString += APRSpadHumid(dataRaw.humidity);
  dataString += "b";              // barometric pressure
  dataString += APRSpadBaro(dataRaw.seaLevelPressure);
  dataString += APRS_DeviceName;
  dataString += " ";
  dataString += APRSver;
  client.println(dataString);     // send to APRS-IS
  Serial.println(dataString);     // print to serial port
  Serial.println("WX Data sent ok");
} // APRSsendWeather()

// *******************************************************
// ************** send telemetry data to APRS ************
// *******************************************************
// page 68 http://www.aprs.org/doc/APRS101.PDF
void APRSsendTelemetry(sensorData dataRaw) {
  String APRSpaddedCall = APRSpadCall(APRScallsign);
  String dataString = APRScallsign;

  // start telemetry messages
  dataString += ">APRS,TCPIP*:";
  dataString += "T#";             // the sequence number must be incremented on each transmission
  dataString += APRSsequence();   // this calls the APRSsequence function
  dataString += ",";
  dataString += APRSpadVcell(dataRaw.cellVoltage);
  dataString += ",";
  dataString += APRSpadRSSI(dataRaw.wifiRSSI);
  dataString += ",";
  dataString += APRSpadLightIntensity(dataRaw.lightIntensity);
  dataString += ",000,000,00000000"; // unused analog & digital channels
  dataString += ",http://W4KRL.com/iot-kits";
  client.println(dataString);        // send to APRS-IS
  Serial.println(dataString);        // print to serial port

  // TO DO: send the  following 3 messages only occaisionaly

  // define parameters
  dataString = APRScallsign;
  dataString += ">APRS,TCPIP*::";
  dataString += APRSpaddedCall;
  dataString += ":PARM.";
  dataString += "Vcell";
  dataString += ",";
  dataString += "RSSI";
  dataString += ",";
  dataString += "Light";
  client.println(dataString);     // send to APRS-IS
  Serial.println(dataString);     // print to serial port

  // provide units for each parameter
  dataString = APRScallsign;
  dataString += ">APRS,TCPIP*::";
  dataString += APRSpaddedCall;
  dataString += ":UNIT.";
  dataString += "Vdc";
  dataString += ",";
  dataString += "dBm";
  dataString += ",";
  dataString += "lux";
  client.println(dataString);     // send to APRS-IS
  Serial.println(dataString);     // print to serial port

  // equations to convert data byte to parameter value
  dataString = APRScallsign;
  dataString += ">APRS,TCPIP*::";
  dataString += APRSpaddedCall;
  dataString += ":EQNS.";
  dataString += "0,0.01961,0";    // convert 0-255 to 0-5V
  dataString += ",";
  dataString += "0,-1,0";         // convert RSSI to dBm
  dataString += ",";
  dataString += "0,257,0";        // convert 0-255 to 0-65535
  client.println(dataString);     // send to APRS-IS
  Serial.println(dataString);     // print to serial port

  // finish telemetry and identify project
  dataString = APRScallsign;
  dataString += ">APRS,TCPIP*::";
  dataString += APRSpaddedCall;
  dataString += ":BITS.";
  dataString += "00000000";
  dataString += ",";
  dataString += "http://w4krl.com/iot-kits";
  client.println(dataString);     // send to APRS-IS
  Serial.println(dataString);     // print to serial port

  Serial.println("Telemetry Report sent ok");
} // APRSsendTelemetry()

// *******************************************************
// ************* logon to APRS-IS server *****************
// *******************************************************
// Thanks to Lei Yang. See https://github.com/wengkai/aprswxbox
boolean wait4APRS(Stream* stream, char *target, int targetLen)
{
  size_t index = 0;  // maximum target string length is 64k bytes!
  int c;
  boolean ret = false;
  unsigned long timeBegin;
  delay(50);
  timeBegin = millis();

  while ( true )
  {
    //  wait and read one byte
    while ( !stream->available() )
    {
      if ( millis() - timeBegin > TO_LINE )
      {
        break;
      }
      delay(2);
    }
    if ( stream->available() ) {
      c = stream->read();
      //  test the byte
      if ( c == target[index] )
      {
        index ++;
        if ( !target[index] )
          // return true if all chars in the target match
        {
          ret = true;
          break;
        }
      }
      else if ( c >= 0 )
      {
        index = 0;  // reset index if any char does not match
      } else //  timed-out for one byte
      {
        break;
      }
    }
    else  //  timed-out
    {
      break;
    }
  }
  return ret;
} // wait4APRS()

// *******************************************************
// ********** Generate Telemetry Sequence Number *********
// *******************************************************
// generate a 3 digit sequence number from 000 to 999
// based on seconds from Jnauray 1, 1900
String APRSsequence() {
  unsigned long seconds = getNtpTime();
  unsigned long interval = seconds / SLEEP_INTERVAL;
  unsigned int valSN = interval % 1000;
  // pad the sequence number to three digits
  String dataString = "";             // clear the string
  if ( valSN < 10 ) {                 // if number is single digit
    dataString += "00";               // pad with two zeros
  }
  else if ( valSN < 100 ) {           // if number is two digits
    dataString += "0";                // pad with one zero
  }
  dataString += valSN;                // put sequence number on right end
  return dataString;
} // APRSsequence()

// *******************************************************
// ***************** Net Time Protocol *******************
// *******************************************************
// cribbed from Tom Igoe and others
const int NTP_PACKET_SIZE = 48;       // NTP time is in the first 48 bytes of message
byte packetBuffer[NTP_PACKET_SIZE];   // buffer to hold incoming & outgoing packets

unsigned long getNtpTime()
{
  IPAddress ntpServerIP;              // NTP server's ip address

  while (udp.parsePacket() > 0) ;     // discard any previously received packets
  Serial.println("Transmit NTP Request");
  // get a random server from the pool
  WiFi.hostByName(ntpServerName, ntpServerIP);
  Serial.print(ntpServerName);
  Serial.print(": ");
  Serial.println(ntpServerIP);
  sendNTPpacket(ntpServerIP);
  uint32_t beginWait = millis();
  while (millis() - beginWait < 1500) {
    int size = udp.parsePacket();
    if (size >= NTP_PACKET_SIZE) {
      Serial.println("Receive NTP Response");
      udp.read(packetBuffer, NTP_PACKET_SIZE);  // read packet into the buffer
      unsigned long secsSince1900;
      // convert four bytes starting at location 40 to a long integer
      secsSince1900 =  (unsigned long)packetBuffer[40] << 24;  // multiple by 16,777,216
      secsSince1900 |= (unsigned long)packetBuffer[41] << 16;  // multiply by 65,536
      secsSince1900 |= (unsigned long)packetBuffer[42] << 8;   // multiply by 256
      secsSince1900 |= (unsigned long)packetBuffer[43];

      return secsSince1900;
    }
  }
  Serial.println("No NTP Response :-(");
  return 0; // return 0 if unable to get the time
} // getNtpTime()

// send an NTP request to the time server at the given address
void sendNTPpacket(IPAddress &address)
{
  // set all bytes in the buffer to 0
  for (int i = 0; i < NTP_PACKET_SIZE; i++) {
    packetBuffer[i] = 0;
  }
  // Initialize values needed to form NTP request
  packetBuffer[0]  = 0b11100011;   // LI, Version, Mode
  packetBuffer[1]  = 0;            // Stratum, or type of clock
  packetBuffer[2]  = 6;            // Polling Interval
  packetBuffer[3]  = 0xEC;         // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12] = 49;
  packetBuffer[13] = 0x4E;
  packetBuffer[14] = 49;
  packetBuffer[15] = 52;
  // all NTP fields have been given values, now
  // you can send a packet requesting a timestamp:
  udp.beginPacket(address, 123);  //NTP requests are to port 123
  udp.write(packetBuffer, NTP_PACKET_SIZE);
  udp.endPacket();
} // sendNTPpacket()

// *******************************************************
// ******************** Blink LED ************************
// *******************************************************
// this flashes the onboard LED to indicate various internal messages
void blinkLED(int flashes)
{
  // must set LED pin to Output in setup()
  for (int i = 0; i < flashes; i++)
  {
    digitalWrite(LEDpin, LOW);  // Turn the LED *ON*
    delay(20);                  // 20 ms for low energy consumption
    digitalWrite(LEDpin, HIGH); // Turn the LED *OFF*
    delay(200);
  }
} // blinkLED()

// *******************************************************
// *********************** END ***************************
// *******************************************************
