/*
   D1M-WX1_IoT-APRS.ino
   D1 Mini Weather Station (Solar)
   Posts to ThingSpeak & APRS-IS

   Set serial monitor to 115,200 baud

   3/14/2017 - release
   Copyright(c) 2017 Berger Engineering

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in all
   copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
*/

#include <Wire.h>                // I2C bus
#include <BME280I2C.h>           // [manager] by Tyler Glenn https://github.com/finitespace/BME280
#include <BH1750.h>              // https://github.com/claws/BH1750
#include <ESP8266WiFi.h>         // ESP8266 WiFi

//#include "config.h"              // Your configuration file
#include "configoffice.h"
// *******************************************************
// ******************* GLOBALS ***************************
// *******************************************************
int LEDpin = 2;                                // for LED on ESP8266 chip
const float HPA_TO_INHG = 0.0295299830714;     // hPa to inHg pressure
const float rExternal = 180000;                // voltage sensor external resistor

// structure to hold sensor measurements in metric units
struct sensorData
{
  float stationPressure;
  float seaLevelPressure;
  float temperature;
  float humidity;
  unsigned int lightIntensity;
  float cellVoltage;
  long wifiRSSI;
} rawData;  // declare struct variables

// used in printToSerial() to select display units
const bool customaryUnits = false;
const bool metricUnits = true;

// *******************************************************
// ******************* DEFAULTS **************************
// *******************************************************

// NTP Server:
static const char ntpServerName[] = "us.pool.ntp.org";

// ********* INSTANTIATE OBJECTS *************************

// Create instances of the sensor objects
BME280I2C myBME280;     // barometric pressure / temperature / humidity sensor
BH1750 myBH1750;        // light intensity sensor

// instantiate the WiFi connection
WiFiClient client;

// *******************************************************
// ********************* SETUP ***************************
// *******************************************************
void setup()
{
  Serial.begin(115200);             // initialize the serial port
  myBME280.begin();                 // initialize BME280 pressure/temperature/humidity
  myBH1750.begin();                 // initialize BH1750 light sensor

  logonToRouter();                  // logon to local Wi-Fi

  rawData = readSensors();          // load all sensor data into rawData struct

  printToSerialPort(rawData, customaryUnits);
  postToThingSpeak(rawData);

  // all done, now go to sleep
  blinkLED(3);                      // notify the user
  Serial.println("Entering deep sleep");
  delay(200);                       // delay to let things settle

  ESP.deepSleep(SLEEP_INTERVAL * 1000000L, WAKE_RF_DEFAULT);
} //setup()

// *******************************************************
// ******************** LOOP *****************************
// *******************************************************
void loop()
{
  // everything is done in Setup()
} // loop()

// *******************************************************
// ****************** SUPPORTING FUNCTIONS ***************
// *******************************************************

// *******************************************************
// ********** read sensors into a sensor struct **********
// *******************************************************
sensorData readSensors()
{
  sensorData tempData;   // temporary variable to hold readings

  // Pressure Parameters: (float pressure, float temperature, float humidity, bool metric, int pressureUnit)
  int pressureUnit = 1;  // unit: 0 = Pa, 1 = hPa, 2 = Hg, 3 = atm, 4 = bar, 5 = torr, 6 = N/m^2, 7 = psi
  myBME280.read(tempData.stationPressure, tempData.temperature, tempData.humidity, true, pressureUnit);

  // calculater the Sea Level Pressure from the station pressure
  float SLP = calculateSeaLevelPressure(tempData.temperature, tempData.stationPressure, STATION_ELEV);
  tempData.seaLevelPressure = SLP;

  // read light level in lux
  tempData.lightIntensity = myBH1750.readLightLevel();

  // read analog voltage from the Analog to Digital Converter
  // on D1 Mini this is 0 - 1023 for voltages 0 to 3.2V
  // the D1M-WX1 has an external resistor to extend the range to 5.0 Volts
  // a fudgeFactor corrects for component variation
  float fudgeFactor = dmmVoltage / adcVoltage;
  tempData.cellVoltage = 5.0 * analogRead(A0) * fudgeFactor / 1023;

  // read the Wi-Fi signal strength
  tempData.wifiRSSI = WiFi.RSSI();

  return tempData;
} // readSensors()

// *******************************************************
// ********* print raw data to the serial port ***********
// *******************************************************
void printToSerialPort(sensorData dataRaw, bool units)
{
  // '\t' is the C++ escape sequence for tab
  // header line
  Serial.println("\tTemp:\tHumd:\tPsta:\tPsea:\tLux:\tVolt:");

  if (units == true) { // display metric units
    // first line - raw data
    Serial.print("RAW\t");
    Serial.print(dataRaw.temperature, 2);
    Serial.print("\t");
    Serial.print(dataRaw.humidity, 2);
    Serial.print("\t");
    Serial.print(dataRaw.stationPressure * HPA_TO_INHG, 2);
    Serial.print("\t");
    Serial.print(dataRaw.seaLevelPressure * HPA_TO_INHG, 2);
    Serial.print("\t");
    Serial.print(dataRaw.lightIntensity);
    Serial.print("\t");
    Serial.println(dataRaw.cellVoltage, 2);
  } else {           // display customary units
    // first line - raw data
    Serial.print("RAW\t");
    Serial.print(dataRaw.temperature * 1.8 + 32, 2);
    Serial.print("\t");
    Serial.print(dataRaw.humidity, 2);
    Serial.print("\t");
    Serial.print(dataRaw.stationPressure, 2);
    Serial.print("\t");
    Serial.print(dataRaw.seaLevelPressure, 2);
    Serial.print("\t");
    Serial.print(dataRaw.lightIntensity);
    Serial.print("\t");
    Serial.println(dataRaw.cellVoltage, 2);
  }
  // second line
  Serial.println("----------------------------------------------------");
} // printToSerialPort()

// *******************************************************
// *********** logon to your Wi-Fi router ****************
// *******************************************************
void logonToRouter()
{
  int count = 0;
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED)
  {
    count++;
    blinkLED(1);
    delay(250);
    Serial.print(".");
    if (count > 15) {
      Serial.print("Exit code: ");
      Serial.println(WiFi.status());
      ESP.reset();
    }
  }
  Serial.print("Connected w/IP: ");
  Serial.println(WiFi.localIP().toString());
} // logonToRouter()


// *******************************************************
// ************* post to ThingSpeak **********************
// *******************************************************
void postToThingSpeak(sensorData data)
{
  // assemble and post the data
  if (client.connect(IOT_SERVER, 80) == true)
  {
    // declare dataString as a String and initialize with the apiWriteKey
    String dataString = apiWriteKey;

    // cocatenate each field onto the end of dataString
    dataString += "&field1=";
    dataString += String(data.temperature);
    dataString += "&field2=";
    dataString += String(data.humidity);
    dataString += "&field3=";
    dataString += String(data.stationPressure);
    dataString += "&field4=";
    dataString += String(data.seaLevelPressure);
    dataString += "&field5=";
    dataString += String(data.lightIntensity);
    dataString += "&field6=";
    dataString += String(data.cellVoltage);
    dataString += "&field7=";
    dataString += String(data.wifiRSSI);
    //    dataString += "&field8=";   // spare field
    //    dataString += String();
    //    dataString += "&status=";   // spare field
    //    dataString += sensorStatus; // unused status field

    // find the number of characters in dataString
    String dataStringLength = String(dataString.length());

    // post the data to ThingSpeak
    client.println("POST /update HTTP/1.1");
    client.println("Host: api.thingspeak.com");
    client.println("Connection: close");
    client.println("X-THINGSPEAKAPIKEY: " + apiWriteKey);
    client.println("Content-Type: application/x-www-form-urlencoded");
    client.println("Content-Length: " + dataStringLength);
    client.println("");
    client.print(dataString);
  }
  client.stop();
} // postToThingSpeak()

// *******************************************************
// calculate relative sealevel pressure from absolute station pressure in hPa
// temperature in °C, elevation in m
// http://www.sandhurstweather.org.uk/barometric.pdf
// http://keisan.casio.com/exec/system/1224575267
// *******************************************************
float calculateSeaLevelPressure(float celsius, float stationPressure, float elevation)
{
  float slP = stationPressure / pow(2.718281828, -(elevation / ((273.15 + celsius) * 29.263)));
  return slP;
} // calculateSeaLevelPressure()

// *******************************************************
// ******************** Blink LED ************************
// *******************************************************
// this flashes the onboard LED to indicate various internal messages
void blinkLED(int flashes)
{
  // must set LED pin to Output in setup()
  for (int i = 0; i < flashes; i++)
  {
    digitalWrite(LEDpin, LOW);  // Turn the LED *ON*
    delay(20);                  // 20 ms for low energy consumption
    digitalWrite(LEDpin, HIGH); // Turn the LED *OFF*
    delay(200);
  }
} // blinkLED()

// *******************************************************
// *********************** END ***************************
// *******************************************************
